/* $Id: db_firebird.c,v 1.3 2005/10/30 12:31:04 zglozik Exp $ */
/* ====================================================================
 * Copyright (c) 2004 Clizio Merli.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 *
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 *
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#include "db_firebird.h"
#include "mod_tsa.h"
#include "tsa_util.h"
#include <openssl/bn.h>
#include <ibase.h>
#include <http_log.h>


/* Global module specific contexts */
#define MOD_TSA_FIREBIRD_PASSWORD       "mod_tsa_firebird_password"
#define MOD_TSA_FB_DEFPASSWD		"masterkey"

/* Default options */
#define DEF_HOST                "localhost"
#define DEF_PORT                3306
#define DEF_UNIX_SOCKET         "/tmp/firebird.sock"
#define DEF_USER                "SYSDBA"
#define DEF_DATABASE            "tsa"
#define DEF_PASSPHRASE          0

/* Configuration object type */
struct DB_firebird_conf_st
{
        char            *host;
        unsigned int    port;
        char            *unix_socket;
        char            *user;
        char            *database;
        int             passphrase;
};

/* Static function declarations */
static int              db_firebird_init1       (server_rec *s);
static DB              *db_firebird_init2       (server_rec *s);
static int              db_firebird_connect     (DB *db, server_rec *s);
static int              db_firebird_insert      (DB *db, request_rec *r,
						 TS_RESP *resp);
static int              db_firebird_disconnect  (DB *db, server_rec *s);
static void             db_firebird_cleanup     (DB *db, server_rec *s);
static unsigned         db_firebird_errno       (DB *db);
static const char      *db_firebird_strerror    (DB *db);

static void             db_clear_errno          (DB *db);
static void             db_hexascii             (char *hexa, char *ascii, 
						 int hexalen);

/* Initialising method table */
DB_methods      db_firebird_methods =
{
        &db_firebird_init1,
        &db_firebird_init2,
        &db_firebird_connect,
        &db_firebird_insert,
        &db_firebird_disconnect,
        &db_firebird_cleanup,
        &db_firebird_errno,
        &db_firebird_strerror
};

/* DB connection state */
struct DB_st
{
        isc_db_handle   db_handle;
        unsigned        db_errno;
        ISC_STATUS      db_status[20];
        long            db_sqlcode;
        char            db_errmsg[256+1];
};

/* Error codes */
#define ERR_BASE                1000
#define ERR_ALLOC_FAILURE       (ERR_BASE + 0)
#define ERR_FIREBIRD_ERROR      (ERR_BASE + 1)

/* Error strings */
static const char *db_errors[] =
{
        "memory allocation failed",
        "firebird db error"
};

/* Configuration handler functions */

extern DB_firebird_conf * db_firebird_create_config
(
        apr_pool_t       *p
)
{
        DB_firebird_conf *c;

        c = (DB_firebird_conf *) apr_pcalloc (p, sizeof(DB_firebird_conf));

        c->host        = DEF_HOST;
        c->port        = DEF_PORT;
        c->unix_socket = DEF_UNIX_SOCKET;
        c->user        = DEF_USER;
        c->database    = DEF_DATABASE;
        c->passphrase  = DEF_PASSPHRASE;

        return c;
}

extern const char * db_firebird_set_host
(
        cmd_parms       *cmd,
        void            *mconfig,
        const char      *host
)
{
        tsa_server_conf *cfg;

        cfg = (tsa_server_conf *) 
		ap_get_module_config (cmd->server->module_config, &tsa_module);

        cfg->db_firebird_conf->host = apr_pstrdup (cmd->pool, host);
        return NULL;
}

extern const char *db_firebird_set_port
(
        cmd_parms       *cmd,
        void            *mconfig,
        const char      *port
)
{
        tsa_server_conf *cfg;

        cfg = (tsa_server_conf *) 
		ap_get_module_config (cmd->server->module_config, &tsa_module);

        cfg->db_firebird_conf->port = atoi (port);
        return NULL;
}

extern const char *db_firebird_set_unix_socket
(
        cmd_parms       *cmd,
        void            *mconfig,
        const char      *socket
)
{
        tsa_server_conf *cfg;

        cfg = (tsa_server_conf *) 
		ap_get_module_config (cmd->server->module_config, &tsa_module);

        cfg->db_firebird_conf->unix_socket = apr_pstrdup (cmd->pool, socket);
        return NULL;
}

extern const char *db_firebird_set_user
(
        cmd_parms       *cmd,
        void            *mconfig,
        const char      *user
)
{
        tsa_server_conf *cfg;

        cfg = (tsa_server_conf *) 
		ap_get_module_config (cmd->server->module_config, &tsa_module);

        cfg->db_firebird_conf->user = apr_pstrdup (cmd->pool, user);
        return NULL;
}

extern const char *db_firebird_set_database
(
        cmd_parms       *cmd,
        void            *mconfig,
        const char      *database
)
{
        tsa_server_conf *cfg;

        cfg = (tsa_server_conf *) 
		ap_get_module_config (cmd->server->module_config, &tsa_module);

        cfg->db_firebird_conf->database = apr_pstrdup (cmd->pool, database);
        return NULL;
}

extern const char *db_firebird_set_passphrase
(
        cmd_parms       *cmd,
        void            *mconfig,
        int             passphrase
)
{
        tsa_server_conf *cfg;

        cfg = (tsa_server_conf *) 
		ap_get_module_config (cmd->server->module_config, &tsa_module);

        cfg->db_firebird_conf->passphrase = passphrase;
        return NULL;
}

/* Functions for handling the database connection */

static int db_firebird_init1
(
        server_rec      *s
)
{
        tsa_server_conf *c;
        char            password[HUGE_STRING_LEN] = "";
        char            message[HUGE_STRING_LEN];

        c = (tsa_server_conf *) 
		ap_get_module_config (s->module_config, &tsa_module);

        /* Ask for the password if required */
        if (c->db_firebird_conf->passphrase)
        {
                snprintf(message, sizeof(message)-1,
                         "Enter passphrase for database %s/%s@%s: ",
                         c->db_firebird_conf->database,
                         c->db_firebird_conf->user,
                         c->db_firebird_conf->host);
                message[sizeof(message)-1] = '\0';
                if (!tsa_get_passphrase(message, password, sizeof(password)))
                {
                        return 0;
                }
                /* Saving password, never released */
                apr_pool_t *p = s->process->pconf;
                apr_pool_userdata_set(apr_pstrdup(p, password), MOD_TSA_FIREBIRD_PASSWORD, 
			NULL, s->process->pool);
        }
	else
        {
                /* Saving password, never released */
                apr_pool_t *p = s->process->pconf;
                apr_pool_userdata_set(apr_pstrdup(p, MOD_TSA_FB_DEFPASSWD), MOD_TSA_FIREBIRD_PASSWORD, 
			NULL, s->process->pool);
        }

        return 1;
}

static DB *db_firebird_init2
(
        server_rec      *s
)
{
        DB              *db;

        db = (DB *) malloc (sizeof(DB));

        if (db == NULL)
        {
                return NULL;
        }

        db_clear_errno (db);

        return db;
}

static int db_firebird_connect
(
        DB              *db,
        server_rec      *s
)
{
        int             ret;
        char            *dpb;
        short           dpb_length;
        char            *passphrase;
        tsa_server_conf *c;

	apr_pool_userdata_get((void **) &passphrase, MOD_TSA_FIREBIRD_PASSWORD, s->process->pool);

        c = (tsa_server_conf *) 
		ap_get_module_config (s->module_config, &tsa_module);

        db_clear_errno (db);

        /* clear database handle */
        db->db_handle = 0L;

        /* add dpb version to dpb */
        dpb = NULL;
        dpb_length = 0;

        /* add user, password, and role to dpb */
        isc_expand_dpb (&dpb, &dpb_length, isc_dpb_user_name,
			c->db_firebird_conf->user,
			isc_dpb_password, passphrase,
			isc_dpb_sql_role_name, "full_access",
			NULL);

        /* attach to database */
        ret = isc_attach_database (db->db_status,
                                   0,
                                   c->db_firebird_conf->database,
                                   &db->db_handle,
                                   dpb_length,
                                   dpb);
        if (ret == 0)
        {
                ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
                             "%s:connection to %s/%s@%s is successful",
                             tsa_module_name,
                             c->db_firebird_conf->database,
                             c->db_firebird_conf->user,
                             c->db_firebird_conf->host);
                ret = 1;
        }
        else
        {
                db->db_sqlcode = -(isc_sqlcode(db->db_status));
                ret = 0;
        }

        return ret;
}

static int db_firebird_insert
(
        DB              *db,
        request_rec     *rqstrec,
        TS_RESP         *resp
)
{
        int             ret;
        isc_tr_handle   isc_trans;
        PKCS7           *token;
        TS_TST_INFO     *tst_info;
        unsigned        token_der_len;
        unsigned char   *pp;
        unsigned char   *token_der;
        ASN1_INTEGER    *serial;
        char            *gen_time;
        BIGNUM          serial_bn;
        char            *serial_hex;
        char            *query;
        char            *esc_token_der;

        ret = 0;
        isc_trans = NULL;

        token    = TS_RESP_get_token (resp);
        tst_info = TS_RESP_get_tst_info (resp);

        pp         = NULL;
        token_der  = NULL;
        serial_hex = NULL;

        query = NULL;

        esc_token_der = NULL;

        db_clear_errno (db);

        BN_init(&serial_bn);

        /* Nothing to do if there is no token in the response, return OK */
        if (token == NULL)
        {
                return 1;
        }

        /* Convert token to DER first */
        token_der_len = i2d_PKCS7 (token, NULL);
        token_der = (unsigned char *) malloc (token_der_len);
        if (token_der == NULL)
        {
                db->db_errno = ERR_ALLOC_FAILURE;
                goto err;
        }
        pp = token_der;
        i2d_PKCS7 (token, &pp);

        /* Escape the DER-encoded token */

        esc_token_der = (char *) apr_palloc (rqstrec->pool, token_der_len*2+1);
        db_hexascii (token_der, esc_token_der, token_der_len);

        /* Get the unique token id as a hex string */
        serial = TS_TST_INFO_get_serial (tst_info);
        if (   !ASN1_INTEGER_to_BN(serial, &serial_bn)
            || !(serial_hex = BN_bn2hex(&serial_bn)))
        {
                db->db_errno = ERR_ALLOC_FAILURE;
                goto err;
        }

        /* Get the time from the token */

        gen_time = ASN1_STRING_data (TS_TST_INFO_get_time(tst_info));

        /* Set up query string */
        query = apr_psprintf (rqstrec->pool,
                             "insert into tsa_events "
                             "(token_id, "
                             " token_date, "
                             " token_pkcs7) "
                             "values ('%s', '%s', '%s')",
                             serial_hex,
                             gen_time,
                             esc_token_der);

        /* Issue the query */

        isc_start_transaction (db->db_status,
                               &isc_trans,
                               1,
                               &db->db_handle,
                               0,
                               NULL);

        ret = isc_dsql_execute_immediate (db->db_status,
                                          &db->db_handle,
                                          &isc_trans,
                                          0,
                                          query,
                                          1,
                                          NULL);
        if (ret != 0)
        {
                db->db_sqlcode = -(isc_sqlcode(db->db_status));
                ret = 0;
                goto err;
        }

        isc_commit_transaction (db->db_status,
                                &isc_trans);

        /* Log the request */

        ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, rqstrec,
                      "%s:new response (serial=0x%s) has been added to DB",
                      tsa_module_name, serial_hex);

        ret = 1;

 err:
        free (token_der);
        BN_free (&serial_bn);
        OPENSSL_free (serial_hex);

        return ret;
}

static int db_firebird_disconnect
(
        DB              *db,
        server_rec      *s
)
{
        db_clear_errno (db);

        isc_detach_database (db->db_status,
                             &db->db_handle);

        return 1;
}

static void db_firebird_cleanup
(
        DB              *db,
        server_rec      *s
)
{
        db_clear_errno (db);
}

extern unsigned db_firebird_errno
(
        DB              *db
)
{
	return db->db_errno != 0 ? db->db_errno : db->db_sqlcode;
}

extern const char *db_firebird_strerror
(
        DB              *db
)
{
        if (db->db_errno != 0)
        {
                return db_errors[db->db_errno-ERR_BASE];
        }

        isc_sql_interprete (-(db->db_sqlcode),
                            db->db_errmsg,
                            sizeof(db->db_errmsg));

        return (db->db_errmsg);
}

static void db_clear_errno
(
        DB              *db
)
{
        db->db_errno   = 0;
        db->db_sqlcode = 0;

        memset (db->db_status, 0, sizeof(db->db_status));
        memset (db->db_errmsg, 0, sizeof(db->db_errmsg));
}

static void db_hexascii
(
        char          * hexa,
        char          * ascii,
        int             hexalen
)
{
        int             i;

        memset (ascii, 0, hexalen*2+1);

        for (i = 0; i < hexalen; ++i)
        {
                sprintf (&ascii[i*2], "%2.2X", hexa[i]);
        }
}
