/* $Id: db_postgresql.c,v 1.1 2006/09/23 10:56:47 zglozik Exp $	*/
/* ====================================================================
 * Copyright (c) 2006 The OpenTSA Project.  All rights reserved.
 *
 * This code is derived from software contributed to OpenTSA project by
 * Tatoku Ogaito.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 *
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 *
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#include "mod_tsa.h"
#include "db_postgresql.h"
#include "tsa_util.h"
#include <openssl/bn.h>
#include <libpq-fe.h>
#include <http_log.h>

#define MOD_TSA_PGSQL_PASSWORD "mod_tsa_pgsql_password"

/* default options */
#define DEF_HOST	"localhost"
#define DEF_PORT	"5432"
#define DEF_USER	"tsa"
#define DEF_DATABASE	"tsa"
#define DEF_PASSPHRASE	0

struct DB_pg_conf_st {
	char *host;
	char *port;
	char *user;
	char *database;
	int   passphrase;
};

static int db_pg_init1(server_rec *);
static DB *db_pg_init2(server_rec *);
static int db_pg_connect(DB *, server_rec *);
static int db_pg_insert(DB *, request_rec *, TS_RESP *);
static int db_pg_disconnect(DB *, server_rec *);
static void db_pg_cleanup(DB *, server_rec *);
static unsigned	int db_pg_errno(DB *);
static const char *db_pg_strerror(DB *);
static int db_clear_errno(DB *);

/* Initialising method table. */
DB_methods db_pg_methods = {
	&db_pg_init1,
	&db_pg_init2,
	&db_pg_connect,
	&db_pg_insert,
	&db_pg_disconnect,
	&db_pg_cleanup,
	&db_pg_errno,
	&db_pg_strerror
};

/* DB connection state. */
struct DB_st
{
	PGconn      *pg_conn;
	unsigned int db_errno;
};

/* Error codes. */
#define	ERR_BASE	1000
#define	ERR_ALLOC_FAILURE	(ERR_BASE + 0)
/* Error strings. */
static const char *db_errors[] =
{
	"memory allocation failed"
};


extern DB_pg_conf *db_pg_create_config(apr_pool_t *p)
{
	DB_pg_conf *c = (DB_pg_conf *)
		apr_pcalloc(p, sizeof(DB_pg_conf));

	c->host = DEF_HOST;
	c->port = DEF_PORT;
	c->user = DEF_USER;
	c->database = DEF_DATABASE;
	c->passphrase = DEF_PASSPHRASE;
	return c;
}

extern const char *db_pg_set_host(cmd_parms *cmd, void *mconfig, const char *host)
{
	tsa_server_conf *cfg = (tsa_server_conf *)
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_pg_conf->host = apr_pstrdup(cmd->pool, host);
	return NULL;
}

extern const char *db_pg_set_port(cmd_parms *cmd, void *mconfig, const char *port)
{
	tsa_server_conf *cfg = (tsa_server_conf *)
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_pg_conf->port = apr_pstrdup(cmd->pool, port);
	return NULL;
}

extern const char *db_pg_set_user(cmd_parms *cmd, void *mconfig, const char *user)
{
	tsa_server_conf *cfg = (tsa_server_conf *)
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_pg_conf->user = apr_pstrdup(cmd->pool, user);
	return NULL;
}

extern const char *db_pg_set_database(cmd_parms *cmd, void *mconfig,
					 const char *database)
{
	tsa_server_conf *cfg = (tsa_server_conf *)
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_pg_conf->database = apr_pstrdup(cmd->pool, database);
	return NULL;
}

extern const char *db_pg_set_passphrase(cmd_parms *cmd, void *mconfig,
					   int passphrase)
{
	tsa_server_conf *cfg = (tsa_server_conf *)
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_pg_conf->passphrase = passphrase;
	return NULL;
}


/*
 *
 */
static int db_pg_init1(server_rec *s)
{
	tsa_server_conf *c;
	char password[HUGE_STRING_LEN];
	char message[HUGE_STRING_LEN];

	bzero(password, sizeof(password));

	c = (tsa_server_conf *)
		ap_get_module_config(s->module_config, &tsa_module);

	if (c->db_pg_conf->passphrase) {
		snprintf(message, sizeof(message)-1,
			 "Enter passphrase for database %s/%s@%s: ",
			 c->db_pg_conf->database,
			 c->db_pg_conf->user,
			 c->db_pg_conf->host);
		message[sizeof(message)-1] = '\0';
		if (!tsa_get_passphrase(message, password, sizeof(password)))
			return 0;

		/* Saving password, never released. */
		apr_pool_t *p = s->process->pconf;
		apr_pool_userdata_set(apr_pstrdup(p, password),
				      MOD_TSA_PGSQL_PASSWORD,
				      NULL, s->process->pool);

	}
	return 1;
}

static DB *db_pg_init2(server_rec *s)
{
	DB *db = (DB *) malloc(sizeof(DB));

	if (!db) return NULL;
	db_clear_errno(db);

	return db;
}


static int db_pg_connect(DB *db, server_rec *s)
{
	char *passphrase;
	tsa_server_conf *c;

	apr_pool_userdata_get((void **) &passphrase, MOD_TSA_PGSQL_PASSWORD,
			      s->process->pool);
	c = (tsa_server_conf *)
		ap_get_module_config(s->module_config, &tsa_module);

	db_clear_errno(db);
	db->pg_conn =  PQsetdbLogin(c->db_pg_conf->host,
				    c->db_pg_conf->port,
				    NULL,
				    NULL,
				    c->db_pg_conf->database,
				    c->db_pg_conf->user,
				    passphrase);
	if (PQstatus(db->pg_conn) == CONNECTION_OK) {
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s: connection to %s/%s@%s is successful",
			     tsa_module_name,
			     c->db_pg_conf->database,
			     c->db_pg_conf->user,
			     c->db_pg_conf->host);
	}

	return 1;
}

static int db_pg_insert(DB *db, request_rec *rec, TS_RESP *resp)
{
	int ret = 0;
	PKCS7 *token = TS_RESP_get_token(resp);
	TS_TST_INFO *tst_info = TS_RESP_get_tst_info(resp);
	size_t token_der_len, esc_token_der_len;
	unsigned char *pp, *token_der = NULL;
	ASN1_INTEGER *serial;
	unsigned char *gen_time;
	BIGNUM serial_bn;
	char *serial_hex = NULL;
	char *query = NULL;
	unsigned char *esc_token_der = NULL;
	PGresult *pg_result = NULL;

	db_clear_errno(db);
	BN_init(&serial_bn);

        /* Nothing to do if there is no token in the response, return OK */
	if(token == NULL)
		return 1;

	/* Convert token to DER first. */
	token_der_len = i2d_PKCS7(token, NULL);
	pp = token_der = (unsigned char *) malloc(token_der_len);
	if (token_der == NULL) {
		db->db_errno = ERR_ALLOC_FAILURE;
		goto err;
	}
	i2d_PKCS7(token, &pp);

	/* Escape the DER-encoded token. */
	esc_token_der = PQescapeByteaConn(db->pg_conn, token_der, token_der_len,
					  &esc_token_der_len);
	if (esc_token_der == NULL)
		goto err;

	/* Get the unique token id as a hex string. */
	serial = TS_TST_INFO_get_serial(tst_info);
	if (!ASN1_INTEGER_to_BN(serial, &serial_bn)
	    || !(serial_hex = BN_bn2hex(&serial_bn))) {
		db->db_errno = ERR_ALLOC_FAILURE;
		goto err1;
	}

	/* Get the time from the token. */
	gen_time = ASN1_STRING_data(TS_TST_INFO_get_time(tst_info));

	/* Set up query string. */
	query = apr_psprintf(rec->pool, "insert into token (token_id, token_date, "
			     "token_pkcs7) values('%s', '%.14s', '%s')",
			     serial_hex, gen_time, esc_token_der);

	/* Issue the query. */
	pg_result = PQexec(db->pg_conn, query);

	if (pg_result == NULL) {
		db->db_errno = ERR_ALLOC_FAILURE;
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, rec,
			      "%s:new response (serial=0x%s) is not recorded to DB: %s",
			      tsa_module_name, serial_hex, PQerrorMessage(db->pg_conn));
		goto err1;
	} else if (PQresultStatus(pg_result) == PGRES_COMMAND_OK) {
		;
	} else {
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, rec,
			      "%s:new response (serial=0x%s) is not recorded to DB: %s",
			      tsa_module_name, serial_hex, PQresultErrorMessage(pg_result));
		goto err2;
	}

	/* Log the request */
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, rec,
                      "%s:new response (serial=0x%s) has been added to DB",
                      tsa_module_name, serial_hex);
        ret = 1;
  err2:
	/* free the storage associated with PGresult */
	PQclear(pg_result);

  err1:
	/* free memory which allocated by PQescabeByteaConn() */
	PQfreemem(esc_token_der);

  err:
	free(token_der);
	BN_free(&serial_bn);
	OPENSSL_free(serial_hex);
	return ret;
}

static int db_pg_disconnect(DB *db, server_rec *rec)
{
	PQfinish(db->pg_conn);
	return 0;
}

static void db_pg_cleanup(DB *db, server_rec *s)
{
	db_clear_errno(db);
	return;
}

static int db_clear_errno(DB *db)
{
	db->db_errno = 0;
	return 0;
}

extern unsigned int db_pg_errno(DB *db)
{
	return db->db_errno != 0 ?
		db->db_errno - ERR_BASE : 0; /*PQerror(db->pg_conn);*/
}

extern const char *db_pg_strerror(DB *db)
{
	return db->db_errno != 0 ?
		db_errors[db->db_errno - ERR_BASE] : PQerrorMessage(db->pg_conn);
}
