/* $Id: tsa_util.c,v 1.12 2006/09/23 19:47:07 zglozik Exp $ */
/* ====================================================================
 * Copyright (c) 2002 The OpenTSA Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 * 
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 * 
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#include "config.h"
#include "tsa_util.h"

#include <unistd.h>
#include <fcntl.h>
#include <http_log.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/pkcs12.h>
#include <openssl/engine.h>

extern const char tsa_module_name[];

/* Server Identification : Modification made by JOUVE <opentsa@jouve-hdi.com>*/
extern char *server_id;

/* Local function declarations. */
static ASN1_INTEGER *next_serial(apr_file_t *serial_fp);
static int save_serial(apr_file_t *serial_fp, ASN1_INTEGER *serial);

#ifdef TS_DEBUG

static char *tsa_error_text = NULL;

extern const char *tsa_error(void)
	{
		return (const char *)tsa_error_text;
	}

#endif

extern void tsa_openssl_init()
	{
	/* Initialising OpenSSL. */
#ifdef TS_DEBUG_MEMORY
	CRYPTO_malloc_debug_init();
	CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);
#endif /* TS_DEBUG_MEMORY */

	OpenSSL_add_all_algorithms();
	ERR_load_ERR_strings();
	ERR_load_crypto_strings();
	ENGINE_load_builtin_engines();
	}

#ifdef TS_DEBUG_MEMORY
static int check_memory_leaks(server_rec *s)
	{
	int success = 0;
	BIO *log = NULL;
	int log_len;
	char *log_data;	/* The data buffer of log. */

	/* Printing leaked memory references if there are any. */
	/* Memory leak check is turned off temporarily, we do not want to
	   report further memory allocations, like the memory for 'log'. */
	MemCheck_off();
	log = BIO_new(BIO_s_mem());
	MemCheck_on();
	if (!log) goto end;
	CRYPTO_mem_leaks(log);
	log_len = (int) BIO_get_mem_data(log, &log_data);
	if (log_len > 0)
		ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
			     "%u: child exit, memory leaks:\n%.*s",
			     (unsigned) getpid(),
			     log_data[log_len-1] == '\n' ? log_len-1 : log_len,
			     log_data);
	else
		ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
			     "%u: child exit, no memory leaks.",
			     (unsigned) getpid());
	
	success = 1;
 end:
	BIO_free_all(log);
	return success;
	}
#endif	/* TS_DEBUG_MEMORY */

extern void tsa_openssl_cleanup(server_rec *s, unsigned long init_pid)
	{
	unsigned long my_pid = (unsigned long) getpid();

	/* Clean up OpenSSL. */
	CONF_modules_unload(1);
	EVP_cleanup();
	ENGINE_cleanup();
	CRYPTO_cleanup_all_ex_data();
	ERR_remove_state(init_pid);
	/* Remove our own error state, too. */
	if (init_pid != my_pid) ERR_remove_state(my_pid);
	EVP_cleanup();
	ERR_free_strings();
#ifdef TS_DEBUG_MEMORY
	check_memory_leaks(s);
#endif /* TS_DEBUG_MEMORY */
	}

/* ----------------------------------------------------------------------
 *	Error handling.
 * ----------------------------------------------------------------------*/

void tsa_print_errors(server_rec *s)
	{
	unsigned long l;
	char buf[256];
	const char *file, *data;
	int line, flags;
	unsigned long es;

	es = CRYPTO_thread_id();
	while ((l = ERR_get_error_line_data(&file, &line, &data, &flags)) != 0)
		{
		ERR_error_string_n(l, buf, sizeof buf);
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:%lu:%s:%s:%d:%s", tsa_module_name,
			     es, buf, file, line,
			     flags & ERR_TXT_STRING ? data : "");
		}
	}

/* ----------------------------------------------------------------------
 *	Certificate/Key loader routines.
 * ----------------------------------------------------------------------*/

X509 *tsa_load_cert(const char *file, int format)
	{
	X509 *x = NULL;
	BIO *cert = NULL;

	if (!(cert = BIO_new_file(file, "r"))) goto end;

	switch (format)
		{
	case FORMAT_ASN1:
		x = d2i_X509_bio(cert, NULL);
		break;
	case FORMAT_PEM:
		x = PEM_read_bio_X509_AUX(cert, NULL, NULL, NULL);
		break;
	case FORMAT_PKCS12:
		{
		PKCS12 *p12 = d2i_PKCS12_bio(cert, NULL);

		PKCS12_parse(p12, NULL, NULL, &x, NULL);
		PKCS12_free(p12);
		p12 = NULL;
		}
		break;
		}
 end:
	BIO_free(cert);
	return x;
	}

STACK_OF(X509) *tsa_load_certs(const char *file)
	{
	int ret = 0;
	BIO *certs;
	int i;
	STACK_OF(X509) *othercerts = NULL;
	STACK_OF(X509_INFO) *allcerts = NULL;

	if (!(certs = BIO_new_file(file, "r"))) goto end;
	if (!(othercerts = sk_X509_new_null())) goto end;

	if (!(allcerts = PEM_X509_INFO_read_bio(certs, NULL, NULL, NULL)))
		goto end;
	for(i = 0; i < sk_X509_INFO_num(allcerts); i++)
		{
		X509_INFO *xi = sk_X509_INFO_value(allcerts, i);
		if (xi->x509)
			{
			sk_X509_push(othercerts, xi->x509);
			xi->x509 = NULL;
			}
		}
	ret = 1;
end:
	if (!ret)
		{
		sk_X509_pop_free(othercerts, X509_free);
		othercerts = NULL;
		}
	sk_X509_INFO_pop_free(allcerts, X509_INFO_free);
	BIO_free(certs);
	return othercerts;
	}

EVP_PKEY *tsa_load_key(const char *file, int format, const char *pass)
	{
	BIO *key = NULL;
	EVP_PKEY *pkey = NULL;

	if (!(key = BIO_new_file(file, "r"))) 
		return NULL;

	switch (format)
		{
	case FORMAT_ASN1:
		pkey = d2i_PrivateKey_bio(key, NULL);
		break;
	case FORMAT_PEM:
		pkey = PEM_read_bio_PrivateKey(key, NULL, NULL, (char *)pass);
		break;
	case FORMAT_PKCS12:
		{
		PKCS12 *p12 = d2i_PKCS12_bio(key, NULL);

		PKCS12_parse(p12, pass, &pkey, NULL, NULL);
		PKCS12_free(p12);
		p12 = NULL;
		}
		break;
		}

	BIO_free(key);
	return pkey;
	}

/*-------------------------------------------------------------------
 * New function to convert ASN1_INTEGER    
 * to string with hexa format
 * Modification made by JOUVE <opentsa@jouve-hdi.com>             
 *------------------------------------------------------------------*/
char *i2s_ASN1_INTEGER_new(X509V3_EXT_METHOD *method, ASN1_INTEGER *a)
{
  BIGNUM *bntmp = NULL;
  char *strtmp = NULL;
  if(!a) return NULL;
  if(!(bntmp = ASN1_INTEGER_to_BN(a, NULL)) ||
     !(strtmp = BN_bn2hex(bntmp)) )
    X509V3err(X509V3_F_I2S_ASN1_INTEGER,ERR_R_MALLOC_FAILURE);
  BN_free(bntmp);
  return strtmp;
}

/*--------------------------------------------------------------------
 * Merge server id and serial number    
 * example : server_id = 0x0F , serial = 2 , result = 0x0F2
 * Modification made by JOUVE <opentsa@jouve-hdi.com>                
 *--------------------------------------------------------------------*/
ASN1_INTEGER *add_server_id(TS_RESP_CTX *ctx,ASN1_INTEGER *serial)
{
  if (!server_id)
	return serial;

  //Convert ASN1_INTEGER TO STRING
  char * serial_str;
  TSA_ERROR_TEXT("tsa_serial_cb: unable to convert ASN1_INTEGER to string");
  if (!(serial_str = i2s_ASN1_INTEGER_new(NULL, serial))) goto err;
  ASN1_INTEGER_free(serial);
  serial = NULL;

  //Merge
  char serverId[1024];
  snprintf(serverId, sizeof(serverId), "%s%s", server_id, serial_str);
  serverId[sizeof(serverId)-1] = '\0';
  OPENSSL_free(serial_str);

  //Convert STRING TO ASN1_INTEGER
  ASN1_INTEGER * new_serial = NULL;
  TSA_ERROR_TEXT("tsa_serial_cb: unable to convert string to ASN1_INTEGER");
  if(!(new_serial=s2i_ASN1_INTEGER(NULL, serverId))) goto err; 
  
  return new_serial;
  
 err:
  
#ifdef TS_DEBUG
  TS_RESP_CTX_set_status_info(ctx, TS_STATUS_REJECTION,
			      tsa_error());
#else
  TS_RESP_CTX_set_status_info(ctx, TS_STATUS_REJECTION,
			      "could not generate serial number");
#endif
  TS_RESP_CTX_add_failure_info(ctx, TS_INFO_ADD_INFO_NOT_AVAILABLE);
  /* Clean up serial number if it was created. */
  ASN1_INTEGER_free(serial);
  return NULL;
}

/* ----------------------------------------------------------------------
 *	Serial number handling routines.
 * ----------------------------------------------------------------------*/

ASN1_INTEGER *tsa_serial_cb(TS_RESP_CTX *ctx, void *data)
	{
	apr_file_t *serial_fp = data;
	ASN1_INTEGER *serial = NULL;

	/* Acquire an exclusive lock for the serial file. */
	TSA_ERROR_TEXT("tsa_serial_cb: unable to lock serial file");
	if (apr_file_lock(serial_fp, APR_FLOCK_EXCLUSIVE) != APR_SUCCESS) goto err;

	TSA_ERROR_TEXT("tsa_serial_cb: no serial");
	serial = next_serial(serial_fp);
	if (!serial) goto err2;

	TSA_ERROR_TEXT("tsa_serial_cb: can't save serial");
	if (!save_serial(serial_fp, serial)) goto err2;

	TSA_ERROR_TEXT("tsa_serial_cb: unable to unlock serial file");
	if (apr_file_unlock(serial_fp) != APR_SUCCESS) goto err;

	TSA_ERROR_TEXT(NULL);
        
        /*********************************************************
         * Merge server id and serial number                     *
         * example : server_id = 0x0F , serial = 2               *
         *           result = 0x0F2                              *
         * Modification made by JOUVE <opentsa@jouve-hdi.com>    *
         *********************************************************/
        ASN1_INTEGER * new_serial = NULL;
	new_serial=add_server_id(ctx,serial);

	return new_serial;
 err2:
	apr_file_unlock(serial_fp);
 err:

#ifdef TS_DEBUG
	TS_RESP_CTX_set_status_info(ctx, TS_STATUS_REJECTION,
				    tsa_error());
#else
	TS_RESP_CTX_set_status_info(ctx, TS_STATUS_REJECTION,
				    "could not generate serial number");
#endif
	TS_RESP_CTX_add_failure_info(ctx, TS_INFO_ADD_INFO_NOT_AVAILABLE);
	/* Clean up serial number if it was created. */
	ASN1_INTEGER_free(serial);
	return NULL;
	}

static ASN1_INTEGER *next_serial(apr_file_t *serial_fp)
	{
	int ret = 0;
	BIO *in = NULL;
	ASN1_INTEGER *serial = NULL;
	BIGNUM *bn = NULL;
	char buf[1024];
	apr_os_file_t filedes;
	apr_os_file_get(&filedes, serial_fp);

	TSA_ERROR_TEXT("next_serial: no new ASN1_INTEGER");
	if (!(serial = ASN1_INTEGER_new())) goto err;

	TSA_ERROR_TEXT("next_serial: no new BIO_fd");
	if (!(in = BIO_new_fd(filedes, BIO_NOCLOSE))) goto err;

	/* Seek to the beginning of file. */
	BIO_seek(in, 0);

	/* Reading last serial number and increase it by one. */
	TSA_ERROR_TEXT("next_serial: no a2i_ASN1_INTEGER");
	if (!a2i_ASN1_INTEGER(in, serial, buf, sizeof(buf))) goto err;

	TSA_ERROR_TEXT("next_serial: no ASN1_INTEGER_to_BN");
	if (!(bn = ASN1_INTEGER_to_BN(serial, NULL))) goto err;

	ASN1_INTEGER_free(serial);
	serial = NULL;

	TSA_ERROR_TEXT("next_serial: no BN_add_word");
	if (!BN_add_word(bn, 1)) goto err;

	TSA_ERROR_TEXT("next_serial: no BN_to_ASN1_INTEGER");
	if (!(serial = BN_to_ASN1_INTEGER(bn, NULL))) goto err;

	TSA_ERROR_TEXT(NULL);
	ret = 1;

 err:
	if (!ret)
		{
		ASN1_INTEGER_free(serial);
		serial = NULL;
		}
	BIO_free_all(in);
	BN_free(bn);
	return serial;
	}

static int save_serial(apr_file_t *serial_fp, ASN1_INTEGER *serial)
	{
	int ret = 0;
	BIO *out = NULL;
	apr_os_file_t filedes;
	apr_os_file_get(&filedes, serial_fp);

	TSA_ERROR_TEXT("save_serial: no BIO_new_fd");
	if (!(out = BIO_new_fd(filedes, BIO_NOCLOSE))) goto err;

	/* Seek to the beginning of file. */
	TSA_ERROR_TEXT("save_serial: no BIO_seek");
	if (BIO_seek(out, 0) < 0) goto err;

	/* Write the data. */
	TSA_ERROR_TEXT("save_serial: no i2a_ASN1_INTEGER");
	if (i2a_ASN1_INTEGER(out, serial) <= 0) goto err;

	TSA_ERROR_TEXT("save_serial: no BIO_puts");
	if (BIO_puts(out, "\n") <= 0) goto err;

	/* Flush and truncate. */
	BIO_flush(out);
	apr_file_trunc(serial_fp, BIO_tell(out));

	TSA_ERROR_TEXT(NULL);
	ret = 1;
 err:
	BIO_free_all(out);
	return ret;
	}

extern int tsa_get_passphrase(const char *message, char *buffer, unsigned size)
	{
	int stderr_fd;	/* To save the old STDERR. */
	int ret = 0;

	/* STDERR is attached to the log file, save and restore it to STDOUT. */
	TSA_ERROR_TEXT("tsa_get_passphrase: dup(STDERR_FILENO)");
	if ((stderr_fd = dup(STDERR_FILENO)) < 0) goto err;
	close(STDERR_FILENO);
	TSA_ERROR_TEXT("tsa_get_passphrase: dup2(STDOUT_FILENO, STDERR_FILENO)");
	if (dup2(STDOUT_FILENO, STDERR_FILENO) < 0) goto err;

	/* Read password from STDERR. */
	TSA_ERROR_TEXT("tsa_get_passphrase: EVP_read_pw_string()");
	if (EVP_read_pw_string(buffer, size, message, 0) != 0) goto err;

	/* Restore old STDERR file descriptor. */
	close(STDERR_FILENO);
	TSA_ERROR_TEXT("tsa_get_passphrase: dup2(stderr_fd, STDERR_FILENO)");
	if (dup2(stderr_fd, STDERR_FILENO) < 0) goto err;
	close(stderr_fd);

	TSA_ERROR_TEXT(NULL);

	ret = 1;
 err:
	return ret;
	}
